# ABSTRACT: base class for tying variables

######################################################################
# Copyright (C) 2021 Asher Gordon <AsDaGo@posteo.net>                #
#                                                                    #
# This program is free software: you can redistribute it and/or      #
# modify it under the terms of the GNU General Public License as     #
# published by the Free Software Foundation, either version 3 of     #
# the License, or (at your option) any later version.                #
#                                                                    #
# This program is distributed in the hope that it will be useful,    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of     #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   #
# General Public License for more details.                           #
#                                                                    #
# You should have received a copy of the GNU General Public License  #
# along with this program. If not, see                               #
# <http://www.gnu.org/licenses/>.                                    #
######################################################################

package Type::Tie::Aggregate::Base;

=head1 DESCRIPTION

This class is inherited by the classes used to tie variables to
types. This class is internal to
L<Type::Tie::Aggregate|Type::Tie::Aggregate>.

=cut

use v5.10.0;
use strict;
use warnings;
use namespace::autoclean;
use Carp;

require Type::Tie::Aggregate::Deep;
*_deep_tie = \&Type::Tie::Aggregate::Deep::deep_tie;

our @CARP_NOT = qw(Type::Tie::Aggregate);

sub new {
    my ($class, $type, @init) = @_;
    my $self = bless {}, $class;
    $self->_type($type);
    $self->initialize(@init);
    return $self;
}

=method initialize

    $obj->initialize(@init);

Initialize C<$obj> from C<@init>. This is the only method that should
be called by non-core code. It should only be called from a variable
tied by C<ttie> like: C<< (tied my %var)->initialize(%init) >>. See
L<Type::Tie::Aggregate/Initialization and Re-initialization> for more
information.

=cut

sub initialize {
    my $self = shift;
    $self->_initialize(@_);
    $self->_check;
    return $self;
}

sub _initialize {
    # It is important that we use a copy for @init (rather than
    # shifting and using @_), because we don't want to modify the
    # original value(s) passed to _initialize().
    my ($self, @init) = @_;
    $self->_ref($self->_create_ref(@init));
}

# Install accessors for 'type' and 'ref', prefixed with an
# underscore. The '_value' accessor, which defaults to the same as
# '_ref', is overridden by Type::Tie::Aggregate::Scalar.
foreach ('type', [ref => 'value']) {
    my ($key, @aliases) = ref eq 'ARRAY' ? @$_ : $_;
    my $code = sub {
	my $self = shift;
	return $self->{$key} unless @_;
	($self->{$key}) = @_;
    };
    foreach ($key, @aliases) {
	no strict 'refs';
	*{"_$_"} = $code;
    }
}

# This is used to check values after coercion. It doesn't do any
# checking by default, but can be overridden by subclasses. It should
# return an error string on error, or undef on success.
sub _check_value { }

# Perform a type check on the type, croaking on error.
# TODO: Optimize this.
sub _check {
    my ($self, %opts) = @_;
    my $type = $self->_type;
    my $old_value = my $value = $self->_value;
    my $retie = ! exists $opts{retie} || $opts{retie};

    if (eval { $type->has_coercion }) {
	if ($type->can('coerce')) {
	    $value = $type->coerce($value);
	    my $err = $self->_check_value($value);
	    croak "Coerced to invalid value: $err" if defined $err;
	    $self->_value($value);
	}
	else {
	    carp "$type falsely reports that it can coerce";
	}
    }

    $type->check($value) or
	croak eval { $type->get_message($value) } //
	"$value did not pass type constraint $type";

    # Retie the type if needed.
    $self->_deep_tie if $retie || $value != $old_value;

    return $value;
}

# Install methods, and also make the methods retie and recheck the
# object if $opts->{mutates}.
sub _install_methods {
    my ($class, $opts, @methods) = @_;

    my $mutates = $opts->{mutates};
    require Type::Tie::Aggregate::Deep if $mutates;

    while (my ($method, $code) = splice @methods, 0, 2) {
	my $callback;
	my $statement; # $code in a single statement
	if (ref $code eq 'CODE') {
	    $callback = $code;
	    $statement = $code = '$self->$callback(@_)';
	}
	else {
	    $statement = "do { $code }";
	}

	if ($callback && ! $mutates) {
	    # Optimize this case.
	    $code = $callback;
	}
	else {
	    $code = q{
		my $self = shift;
	    } . (
		! $callback || $mutates ? q{
		    my $ref = $self->_ref;
		} : '',
	    ) . (
		$mutates ? qq{
		    wantarray ? my \@ret = $statement :
			defined wantarray ? my \$ret = $statement :
			$statement;
		} . q{
		    $self->_check;
		    return wantarray ? @ret : $ret;
		} : $code,
	    );

	    $code = eval "sub { $code }" or die;
	}

	no strict 'refs';
	*{"$class\::$method"} = $code;
    }
}

=head1 SEE ALSO

=for :list
* L<Type::Tie::Aggregate>

=cut

1;
